"""
Handles information from the FSE airport database
"""
import os

import pandas as pd
from geographiclib.geodesic import Geodesic

from pyfseconomy.utils import GeoLocation

DATA_LOCATION = os.path.join(os.path.dirname(__file__), "icaodata.csv")


class Airports:
    """
    Class of methods that allow retrieval of useful information about airports.
    """
    METERS_PER_NM = 1852
    _airport_data = pd.read_csv(DATA_LOCATION)

    @classmethod
    def get_airport_distance(cls, departure_icao, arrival_icao):
        """
        Calculate the distance in nautical miles between two airports

        :param departure_icao: ICAO code of departure airport
        :param arrival_icao: ICAO code of arrival airport
        :return: Distance in NM between the two airports
        """
        distance_nm, _ = cls.get_airport_distance_bearing(departure_icao, arrival_icao)
        return distance_nm

    @classmethod
    def get_airport_bearing(cls, departure_icao, arrival_icao):
        """
        Calculate the bearing in degrees going from departure to arrival airport

        :param departure_icao: ICAO of departure airport
        :param arrival_icao: ICAO of arrival airport
        :return: Bearing in degrees to travel from dep to arr airport
        """
        _, bearing = cls.get_airport_distance_bearing(departure_icao, arrival_icao)
        return bearing

    @classmethod
    def get_airport_distance_bearing(cls, departure_icao, arrival_icao):
        """
        Calculate the distance and bearing between the departure and arrival

        :param departure_icao: ICAO of departure airport
        :param arrival_icao: ICAO of arrival airport
        :return: Distance in nautical miles and bearing
        """
        airport_data = cls._airport_data.set_index("icao")

        result = Geodesic.WGS84.Inverse(airport_data.at[departure_icao, 'lat'], airport_data.at[departure_icao, 'lon'],
                                        airport_data.at[arrival_icao, 'lat'], airport_data.at[arrival_icao, 'lon'])
        distance_nm = result['s12'] / cls.METERS_PER_NM
        bearing_def = result['azi1']

        if bearing_def < 0:
            bearing_def = 360 + bearing_def

        return distance_nm, bearing_def

    @classmethod
    def get_airports_within(cls, icao, max_distance, min_distance=0, civil_only=True):
        """
        Report a list of airports within a radius

        :param civil_only:
        :param min_distance:
        :param icao: Airport at centre
        :param max_distance: Search radius - in nautical miles
        :return: Dataframe of airports within radius, giving their distance and heading
        """
        airport_data_by_icao = cls._airport_data.set_index("icao")
        max_distance_km = max_distance * cls.METERS_PER_NM / 1000

        # get the bounding coordinates for the radius
        airport_location = GeoLocation(airport_data_by_icao.at[icao, 'lat'], airport_data_by_icao.at[icao, 'lon'])
        min_lat, min_lon, max_lat, max_lon = airport_location.bounding_coordinates(max_distance_km)

        # get locations within the bounding area, exclude self!
        locations = cls._airport_data[(min_lon <= cls._airport_data['lon']) & (cls._airport_data['lon'] <= max_lon) &
                                      (min_lat <= cls._airport_data['lat']) & (cls._airport_data['lat'] <= max_lat) &
                                      (icao != cls._airport_data['icao'])]
        if civil_only:
            locations = locations[(locations['type'] == 'civil')]

        locations.set_index('icao', inplace=True)
        icao_list = pd.DataFrame(columns=['icao', 'distance', 'bearing'])
        for location_icao, location in locations.iterrows():
            distance, bearing = cls.get_airport_distance_bearing(icao, location_icao)
            if min_distance < distance <= max_distance:
                df = pd.DataFrame([[location_icao, distance, bearing]], columns=icao_list.columns)
                icao_list = icao_list.append(df)

        icao_list.set_index('icao', inplace=True)
        return icao_list



"""
Aircraft data
"""
from enum import Enum, auto


class AircraftTypes(Enum):
    """
    Enumeration of aircraft and associated FSEconomy alias strings
    """
    TBM_930 = auto()
    KING_AIR_350 = auto()
    C172_SKYHAWK = auto()
    C152_AEROBAT = auto()
    CITATION_X = auto()
    MSFS_A320 = auto()
    BOEING_737_800 = auto()
    BOEING_747_400 = auto()
    CESSNA_GRAND_CARAVAN = auto()

    @staticmethod
    def list():
        """
        Generate a list strings for all aircraft supported
        :return: List of all aircraft in the enumeration by their strings
        """
        return list(map(lambda a: str(a), AircraftTypes))

    def __str__(self):
        """
        Convert enumeration into FSE alias string
        :return: FSE alias string
        """
        if self.value == self.TBM_930.value:
            return "Socata TBM 930 (MSFS)"
        elif self.value == self.KING_AIR_350.value:
            return "Beechcraft King Air 350"
        elif self.value == self.C172_SKYHAWK.value:
            return "Cessna 172 Skyhawk"
        elif self.value == self.C152_AEROBAT.value:
            return "Cessna 152 Aerobat"
        elif self.value == self.CITATION_X.value:
            return "Cessna Citation X"
        elif self.value == self.MSFS_A320.value:
            return "Airbus A320 (MSFS)"
        elif self.value == self.BOEING_737_800.value:
            return "Boeing 737-800"
        elif self.value == self.BOEING_747_400.value:
            return "Boeing 747-400"
        elif self.value == self.CESSNA_GRAND_CARAVAN.value:
            return "Cessna 208 Caravan"
        else:
            return f"Unknown plane type - {self.name}"


class Aircraft:
    """
    Hold information relating to an aircraft
    """
    FUEL_KG_PER_GAL = 2.687344961
    PAX_WEIGHT_KG = 77

    def __init__(self, make_model, mtow, seats, crew, empty_weight, fuel_total):
        """

        :param make_model: Make and model string
        :param mtow: Maximum Takeoff Weight
        :param seats: Number of available seats
        :param crew: Number of required crew
        :param empty_weight: Empty weight
        :param fuel_total: Amount of fuel the plane is capable of carrying in total
        """
        self._name = make_model
        self._mtow = mtow
        self._seats = seats
        self._crew = crew
        self._empty_weight = empty_weight
        self._fuel_total_gal = fuel_total

    def get_name(self):
        """
        Get aircraft name

        :return: Name of aircraft
        """
        return self._name

    def get_max_pax(self):
        """
        Get maximum number of passengers

        :return: Number of passengers (pax = seats - crew - pilot)
        """
        return self._seats - self._crew - 1

    def get_pax_cargo_for_fuel(self, fuel_percent):
        """
        Return the amount of pax that can be carried given the fuel load in % or the total amount of cargo that can be
        carried.

        :param fuel_percent: Percentage total fuel load for the aircraft
        :return: Number of pax with onboard fuel and the maximum total available payload (including pax) for the fuel
        """
        fuel_weight = fuel_percent * self._fuel_total_gal * self.FUEL_KG_PER_GAL
        available_payload = self._mtow - self._empty_weight - fuel_weight
        available_pax = min(self.get_max_pax(), int(available_payload / self.PAX_WEIGHT_KG))

        return available_pax, available_payload

    @classmethod
    def aircraft_from_data(cls, csv_data):
        """
        Create an instance of an aircraft based on information from the FSE CSV.
        :param csv_data: Pandas CSV dataframe created from FSE CSV
        :return: Aircraft class
        """
        fuel_total = csv_data['Ext1'] + csv_data['LTip'] + csv_data['LAux'] + csv_data['LMain']
        fuel_total += csv_data['Center1'] + csv_data['Center2'] + csv_data['Center3']
        fuel_total += csv_data['RExt2'] + csv_data['RTip'] + csv_data['RAux'] + csv_data['RMain']

        return cls(csv_data['MakeModel'], csv_data['MTOW'], csv_data['Seats'], csv_data['Crew'],
                   csv_data['EmptyWeight'], fuel_total)



"""
This file contains useful functions that can be used to process data gathered from FSE data feeds via the client
"""
import math
from collections import OrderedDict
from operator import itemgetter

from pyfseconomy import Aircraft, TripTypes


class PlaneLoadingError(BaseException):
    pass


def order_assignment_by_value(assignments):
    """
    Calculate the value - the $/kg - of each of the assignments, then sort them by value into a dictionary keyed
    with the assignment ID

    :param assignments: Assigments dataframe
    :return: OrderedDict of assignments, keyed by ID - sorted by the value which is $/kg.
    """
    assignment_value_list = {}
    for _, assignment in assignments.iterrows():
        pay = assignment['Pay']
        weight = assignment['Amount']
        if pay > 0 and weight > 0:
            if assignment['UnitType'] == 'passengers':
                weight *= Aircraft.PAX_WEIGHT_KG
            value = pay / weight
            assignment_value_list[assignment['Id']] = value
    return OrderedDict(sorted(assignment_value_list.items(), key=itemgetter(1), reverse=True))


def load_assignments(assignment_list, assignments_by_id, intial_remaining_payload, intial_remaining_pax, aircraft_id):
    """
    Loads assignments up to the remaining payload or pax in the order given.

    :param assignment_list: A list of assignments in the order in which they should be loaded.
    :param assignments_by_id: Dataframe of assignments indexed by their FSE ID
    :param intial_remaining_payload: Starting available payload that can be used (include Pax in this)
    :param intial_remaining_pax: Starting available pax that can be used
    :param aircraft_id: FSE Aircraft identifier - used to load All In trips that are tied to a specific aircraft
    :return: Tuple of Total value of assignments and list of loaded assignment IDs
    """
    assignment_ids = []
    remaining_payload = intial_remaining_payload
    remaining_pax = intial_remaining_pax
    total_value = 0

    for assignment_id, _ in assignment_list.items():
        load_into_plane = True

        unit_type = assignments_by_id.at[assignment_id, "UnitType"]
        amount = assignments_by_id.at[assignment_id, "Amount"]
        pay = assignments_by_id.at[assignment_id, "Pay"]
        assignment_name = f"{amount} {assignments_by_id.at[assignment_id, 'Commodity']}"
        type_is_vip = assignments_by_id.at[assignment_id, "Type"] == str(TripTypes.VIP)
        type_is_allin = assignments_by_id.at[assignment_id, "Type"] == str(TripTypes.ALL_IN)
        exclusive_load = type_is_allin or type_is_vip

        if exclusive_load and remaining_payload != intial_remaining_payload:
            print("WARNING: Attempted to load VIP or All-In job with normal Trip payload")
            continue

        if assignments_by_id.at[assignment_id, "Type"] == str(TripTypes.ALL_IN):
            # Only load if it's associated with this requested_plane
            load_into_plane = aircraft_id == assignments_by_id.at[assignment_id, "AircraftId"]

        if unit_type == 'passengers':
            if amount > remaining_pax:
                load_into_plane = False
            amount *= Aircraft.PAX_WEIGHT_KG

        if amount > remaining_payload:
            load_into_plane = False

        if load_into_plane:
            assignment_ids.append((assignment_id, assignment_name))
            total_value += pay
            remaining_payload -= amount
            if unit_type == 'passengers':
                remaining_pax -= amount / Aircraft.PAX_WEIGHT_KG

        # no more jobs - we now have 1 exclusive payload
        if exclusive_load and load_into_plane:
            break

    return total_value, assignment_ids


def plane_loader(assignments, aircraft_id, max_payload=None, max_pax=None, fuel_pct=1.0,
                 aircraft_data=None):
    """
    Loads the most valuable assignments first until one doesn't fit and then continues to load the next most valuable
    that will fit up until the max_payload or max_pax is reached, but not exceeded.

    Will also only load planes with a single VIP or All in load.

    :param aircraft_id: ID/Serial number of the aircraft (not the registration)
    :param aircraft_data: Aircraft class populated with information about the aircraft to load
    :param fuel_pct: Percentage fuel that is in the requested_plane to be loaded.
    :param assignments: Dataframe containing full data for assignments in the assignment_list
    :param max_payload: Max total payload (including pax) the requested_plane and hold at the fuel level you want, set
                        to None for automatic values
    :param max_pax: Maximum number of passengers
    :return: Total value and list of loaded assignments IDs
    """
    assignments_by_id = assignments.set_index("Id")
    if aircraft_data is not None:
        starting_pax, starting_payload = aircraft_data.get_pax_cargo_for_fuel(fuel_pct)
    elif max_payload is not None and max_pax is not None:
        starting_payload = max_payload
        starting_pax = max_pax
    else:
        raise PlaneLoadingError("Expect max_payload and max_pax or aircraft_data")

    assignment_values = order_assignment_by_value(assignments)
    total_value, assignment_ids = load_assignments(assignment_values, assignments_by_id, starting_payload, starting_pax,
                                                   aircraft_id)

    return total_value, assignment_ids


class GeoLocation:
    """
    Porting of functions from http://janmatuschek.de/LatitudeLongitudeBoundingCoordinates#Java

    This assumes a sphere.
    """

    MIN_LAT = -math.pi / 2
    MAX_LAT = math.pi / 2
    MIN_LON = -math.pi
    MAX_LON = math.pi
    EARTH_RADIUS_KM = 6371.01

    def __init__(self, lat, lon):
        self.rad_lat = math.radians(lat)
        self.rad_lon = math.radians(lon)

    def bounding_coordinates(self, distance, radius=EARTH_RADIUS_KM):
        """
        Calculate the max latitude and longitude a distance from a point on a sphere

        :param distance: Max distance from the point, in KM
        :param radius: Radius of the sphere
        :return: min latitude, min longitude, max latitude, max longitude
        """

        # angular distance in radians on a great circle
        rad_dist = distance / radius

        min_lat = self.rad_lat - rad_dist
        max_lat = self.rad_lat + rad_dist

        # calculate the min & max longitude
        if min_lat > self.MIN_LAT and max_lat < self.MAX_LAT:
            delta_lon = math.asin(math.sin(rad_dist) / math.cos(self.rad_lat))

            min_lon = self.rad_lon - delta_lon
            if min_lon < self.MIN_LON:
                min_lon += 2 * math.pi

            max_lon = self.rad_lon + delta_lon
            if max_lon > self.MAX_LON:
                max_lon -= 2 * math.pi
        else:
            # pole is within bounding area
            min_lat = max(min_lat, self.MIN_LAT)
            max_lat = min(max_lat, self.MAX_LAT)
            min_lon = self.MIN_LON
            max_lon = self.MAX_LON

        return math.degrees(min_lat), math.degrees(min_lon), math.degrees(max_lat), math.degrees(max_lon)
